/* ------------------------------------------------------------------------
   file:   gaptrack.c - All important program logic
   description: This file contains all the code to locate the silence, and
   split the wav on the silence boundaries.
   ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sndfile.h>
#include "cmdline.h"
#include "linked_list.h"

/* number of buffer bytes is actually 2x this number */
#define BUFFER_LEN (1024*32)
#define FILENAME_LEN 1024

/** represents a segment of silence */
typedef struct _SILENCE {
	int sample_midpoint; /**< midpoint of the silence, in number of samples into the wav file */
	int end_block;       /**< TRUE if this is the last SILENCE struct in the list */
} SILENCE;

/**
 * Adds SILENCE instances to the silence_list based on input from the file infilename.
 */
int find_silence(int volthreshold, int intervaltime, const char* infilename, linked_list silence_list, int verbose);
/**
 * Does the actual splitting of the sound file into pieces
 */
int generate_tracks(int shortest, const char* infilename, linked_list silence_list, const char* outdir, int verbose);
/**
 * Used by generate tracks to get the output filename
 */
int get_next_filename(char* outfilename, const char* infilename, int segment_number, const char* outdir);
/**
 * Gets the last token in a string delimited by delim
 */
void get_last_token(char* dest, const char* tokenize, const char* delim);

/**
 * Main entry point
 */
int main (int argc, char *argv[]) {
	char *infilename, *outdir;
	/* the volume at which fuzz becomes sound */
	int volthreshold;
	/* the minimum period of silence between two tracks */
	int intervaltime;
	/* spam the user with useless info */
	int verbose;
	/* minimum length of the output tracks */
	int mintracklen;
	
	int errored;
	linked_list silence_list;

	silence_list = list_create();
	cmdline(argc, argv, &infilename, &outdir, &volthreshold, &intervaltime, &mintracklen, &verbose);
	printf("Volume threshold: %d\n", volthreshold);
	printf("Minimum inter-track interval: %ds\n", intervaltime);
	printf("Minimum track length: %ds\n", mintracklen);
	printf("Output directory: \"%s\"\n", outdir);
	if (verbose) {
		printf("Verbosity: on\n");
	}
	printf("\n");

	errored = FALSE;

	if (!find_silence(volthreshold, intervaltime, infilename, silence_list, verbose)) {
		printf("Error during silence detection.\n");
		errored = TRUE;
	}

	if (!errored &&
		!generate_tracks(mintracklen, infilename, silence_list, outdir, verbose)) {
		printf("Error generating tracks.\n");
		errored = TRUE;
	}

	/* free the elements of the silence list */
	list_enumerate(silence_list, (listcallback*)&free);
	/* destroy the silence list */
	list_destroy(silence_list);

	return errored;
} /* main */


int find_silence(int volthreshold, int intervaltime, const char* infilename, linked_list silence_list, int verbose) {
	
	short data [BUFFER_LEN];
	SNDFILE	*infile;
	SF_INFO	sfinfo;
	unsigned int readcount, len;
	int samples_below_thresh, samples_needed_for_interval, cur_sample_number;

	if (verbose) {
		printf("Begining silence detection...\n");
	}
	
	if (!(infile = sf_open (infilename, SFM_READ, &sfinfo))) {
		printf ("Unable to open input file \"%s\".\n", infilename);
		sf_perror (NULL);
		return FALSE;
	}

	len = BUFFER_LEN - (BUFFER_LEN % sfinfo.channels);

	samples_needed_for_interval = sfinfo.samplerate * intervaltime;
	/* printf("samples needed for interval: %d\n", samples_needed_for_interval); */
	
	cur_sample_number = 0;
	samples_below_thresh = 0;
	while ((readcount = sf_read_short (infile, data, len))) {	
		int samples_in_buffer;
		int i;
		
		samples_in_buffer = readcount / sfinfo.channels;
		for (i=0; i<samples_in_buffer; i++) {
			short* cur_sample;
			int channel;
			int below_thresh;
			
			cur_sample = &data[i * sfinfo.channels];
			
			/* find out if this sample is below the threshhold */
			below_thresh = 1;
			for (channel = 0; channel < sfinfo.channels; channel++) {
				/* printf("sample[%d][%d] = %d\n", cur_sample_number, channel, cur_sample[channel]); */
				if (labs(cur_sample[channel]) >= volthreshold) {
					below_thresh = 0;
					break;
				} 
			}
			
			if (below_thresh) {
				/* if the sample was below the threshhold, then increase the counter */
				samples_below_thresh++;
			} else {
				/* if the sample was not below the threshhold, then see how long we were below the thresh */
				
				if (samples_below_thresh >= samples_needed_for_interval) {
					/* we were below the thresh for long enough to count the interval */
					SILENCE* silence = (SILENCE*)malloc(sizeof(SILENCE));
					int sample_start, sample_end;

					sample_start = cur_sample_number - samples_below_thresh;
					sample_end = cur_sample_number;

					/* divide by sfinfo.samplerate to get seconds in */
					silence->sample_midpoint = (sample_start + sample_end) / 2;
					silence->end_block = FALSE;

					if (verbose) {
						printf("Found silence at %ds (%d) to %ds (%d)\n", sample_start / sfinfo.samplerate, sample_start, sample_end / sfinfo.samplerate, sample_end);
					}					

					list_add_tail(silence_list, silence);
				}
				samples_below_thresh = 0;
			}
			
			/* increment the sample number we are on */
			cur_sample_number++;
		}
	}

	sf_close (infile);

	{
		SILENCE* end_block;
		end_block = (SILENCE*)malloc(sizeof(SILENCE));
		end_block->sample_midpoint = cur_sample_number;
		end_block->end_block = TRUE;
		list_add_tail(silence_list, end_block);
	}

	
	return TRUE;
}

void get_last_token(char* dest, const char* tokenize, const char* delim) {
	char tokenstr[FILENAME_LEN], *cur, *last;
	
	strncpy(tokenstr, tokenize, FILENAME_LEN);
		
	last = tokenstr;
	cur = strtok(tokenstr, delim);
	while ((cur = strtok(NULL, delim))) {
		last = cur;
	}

	strncpy(dest, last, FILENAME_LEN);
}

int get_next_filename(char* outfilename, const char* infilename, int segment_number, const char* outdir) {
	int dirlen, outlen, suffixlen, slashlen, prefixlen, extensionlen;
	char prefix[FILENAME_LEN];
	char extension[FILENAME_LEN];
	char slash[2];
	char suffix[32];

	/* figure out the prefix and extension */
	get_last_token(prefix, infilename, "/");
	get_last_token(extension, prefix, ".");
	
	prefixlen = strlen(prefix);
	extensionlen = strlen(extension);
	
	if (prefixlen != extensionlen) {
		/* if they are not the same string, then subtract the extension from the prefix */ 
		prefixlen -= extensionlen + 1;
		prefix[prefixlen] = 0;
	} else {
		/* if they are the same string, then there is no extension */
		extension[0] = 0;
		extensionlen = 0;
	}

	/* do slash */
	slash[0] = '/';
	slash[1] = 0;

	/* figure out the suffix */
	if (extensionlen == 0) {
		sprintf(suffix, "[%04d]", segment_number);
	} else {
		sprintf(suffix, "[%04d].", segment_number);
	}

	dirlen = strlen(outdir);
	slashlen = strlen(slash);
	suffixlen = strlen(suffix);

	outlen = dirlen + slashlen + prefixlen + suffixlen + extensionlen + 1;

	if (outlen > FILENAME_LEN) {
		return FALSE;
	}

	sprintf(outfilename, "%s%s%s%s%s", outdir, slash, prefix, suffix, extension);
	
	return TRUE;
}

int generate_tracks(int shortest, const char* infilename, linked_list silence_list, const char* outdir, int verbose) {
	SNDFILE	*infile, *outfile;
	SF_INFO	sfinfo, sfoutinfo;
	int current_track, current_frame, last_track_end;
	char outfilename[FILENAME_LEN];
	SILENCE* silence;

	if (verbose) {
		printf("Begining track generation...\n");
	}
	
	if (!(infile = sf_open(infilename, SFM_READ, &sfinfo))) {
		printf("Unable to open input file \"%s\".\n", infilename);
		sf_perror(NULL);
		return FALSE;
	}

	memcpy(&sfoutinfo, &sfinfo, sizeof(SF_INFO));
	sfoutinfo.frames = 0;

	current_track  = 0;
	current_frame = 0;
	last_track_end = 0;

	while ((silence = (SILENCE*)list_remove_head(silence_list))) {
		int remaining_frames, channels_to_read;
		unsigned int len;
		short data[BUFFER_LEN]; 

		if ((silence->sample_midpoint - last_track_end) / sfinfo.samplerate < shortest &&
			!silence->end_block) {
			free(silence);
			continue;
		}

		len = BUFFER_LEN - (BUFFER_LEN % sfinfo.channels);

		if (!get_next_filename(outfilename, infilename, current_track, outdir)) {
			printf("Error generating output filename.\n");
			sf_close (infile);
			free(silence);
			return FALSE;
		}
		
		if (verbose) {
			printf("%s\n", outfilename);
		}

		if (!(outfile = sf_open(outfilename, SFM_WRITE, &sfoutinfo))) {
			printf("Unable to open output file \"%s\".\n", outfilename);
			sf_perror(NULL);
			sf_close(infile);
			free(silence);
			return FALSE;
		}

		remaining_frames = silence->sample_midpoint - current_frame;
		
		while ((channels_to_read = (len < remaining_frames*sfinfo.channels) ? len : remaining_frames*sfinfo.channels)) {
			unsigned int readcount, writecount;

			readcount = sf_read_short(infile, data, channels_to_read);
			if (readcount != channels_to_read) {
				printf("Unable to read samples from input file \"%s\".\n", infilename);
				sf_perror(infile);
				sf_close(infile);
				sf_close(outfile);
				free(silence);
				return FALSE;
			}

			writecount = sf_write_short(outfile, data, readcount);
			if (writecount != channels_to_read) {
				printf("Unable to write samples to output file \"%s\".\n", outfilename);
				sf_perror(outfile);
				sf_close(infile);
				sf_close(outfile);
				free(silence);
				return FALSE;
			}
			
			remaining_frames -= channels_to_read/sfinfo.channels;
			current_frame += channels_to_read/sfinfo.channels;
		}

		sf_close (outfile);
		free(silence);
		current_track++;
		last_track_end = current_frame;
	}
		
	sf_close (infile);

	return TRUE;
}



