/* ------------------------------------------------------------------------
   file:   linked_list.c - Linked List ADT
   description: This is a linked list I made in college.
   ------------------------------------------------------------------------ */

#include "linked_list.h"
#include <assert.h>
#include <stdio.h>
#include <memory.h>
#include <stdlib.h>

/*#ifdef _DEBUG*/
#if 0
#define ENSURE_VALID(list) list_valid(list, __FUNCTION__, __LINE__);

static void list_valid(linked_list list, char* func, int line) 
{
  list_node* walker = &list->root;    

  while (walker) {                   
    if (walker->magic!=0xFEEDFACE){   
      printf("List corruption detected. During call to %s (line %d) (last list call was %s)\n", func, line, list->lastfunc?list->lastfunc:"none");
      exit(-1);
    }
    walker = walker->next;
  }

  if (list->current && list->current->magic!=0xFEEDFACE){
    printf("List corruption detected. During call to %s (line %d) (last list call was %s)\n", func, line, list->lastfunc?list->lastfunc:"none");
    exit(-1);
  }

  list->lastfunc = func;
}

#else 
#define ENSURE_VALID(list)
#endif 


/* -------------------------------------------------------------------------- */
/* construct a list */
linked_list list_create(){
  linked_list newlist = (linked_list) malloc(sizeof(_llist));
  memset(newlist, 0, sizeof(_llist));
  newlist->root.magic = 0xFEEDFACE;
  ENSURE_VALID(newlist);
  return newlist;
}

/* -------------------------------------------------------------------------- */
/* destroy a list */
void list_destroy(linked_list list){
  ENSURE_VALID(list);
  
  list_clear(list);

  /* free the list */
  free(list);
}

/* -------------------------------------------------------------------------- */
/* clear the list */
void list_clear(linked_list list){
  list_node *walker, *last;
  ENSURE_VALID(list);


  walker = list->root.next;

  /* free each node */
  while (walker){
    last = walker;
    walker = walker->next;

    free(last);
  }

  list->root.next = NULL;
  list->current = NULL;
  list->tail = NULL;

  list->count = 0;

  ENSURE_VALID(list);
}

/* -------------------------------------------------------------------------- */
/* for internal use, creates a node */
list_node* list_create_node(void *data){
  list_node *newnode = (list_node*) malloc(sizeof(list_node));
  newnode->next = NULL;
  newnode->data = data;
  newnode->magic = 0xFEEDFACE;

  return newnode;
}

/* -------------------------------------------------------------------------- */
/* add a data element to the end of the list*/
void list_add_tail(linked_list list, void *data){
  list_node *newnode;
  ENSURE_VALID(list);
  newnode = list_create_node(data);

  /* if there is a tail, update it */
  if (list->tail)
    list->tail->next = newnode;
  /* if theres no tail, theres also no head, update it */
  else
    list->root.next = newnode;
  list->tail = newnode;
  
  list->count++;
  ENSURE_VALID(list);
}

/* -------------------------------------------------------------------------- */
/* add a data element to the front of the list */
void list_add_head(linked_list list, void *data){
  list_node *newnode;
  ENSURE_VALID(list);
  newnode = list_create_node(data);

  newnode->next = list->root.next;
  list->root.next = newnode;

  if (!list->tail)
    list->tail = newnode;

  list->count++;
  ENSURE_VALID(list);
}

/* -------------------------------------------------------------------------- */
/* returns TRUE if the element exists in the list */
int list_contains(linked_list list, void* data)
{
  list_node *walker;
  ENSURE_VALID(list);

  walker = list->root.next;
  
  while (walker){
    if (walker->data == data){
      return TRUE;
    }
    walker = walker->next;
  }
  
  return FALSE;
}

/* -------------------------------------------------------------------------- */
/* removes an element from the list, returns TRUE
   if the element existed in the list*/
int list_remove(linked_list list, void* data)
{
  list_node *walker, *last;
  ENSURE_VALID(list);

  walker = &list->root;

  while (walker){
    last = walker;
    walker = walker->next;

    if (walker && walker->data == data){
      last->next = walker->next;
      if (walker == list->tail)
	list->tail = last;

      /* if (walker == list->current) */
	list->current=walker->next;
      
      free(walker);
      list->count--;
      
      ENSURE_VALID(list);
      return TRUE;
    }
  }
  
  return FALSE;
}

/* -------------------------------------------------------------------------- */
/* remove the head of the list and returns it */
void* list_remove_head(linked_list list)
{
  void* data;
  ENSURE_VALID(list);

  if (!list->root.next)
    return NULL;
  
  data = list->root.next->data;
  list_remove(list, list->root.next->data);
  
  ENSURE_VALID(list);
  return data;
}

/* -------------------------------------------------------------------------- */
/* remove the tail of the list and returns it */
void* list_remove_tail(linked_list list)
{
  void *data;
  ENSURE_VALID(list);

  if (!list->root.next)
    return NULL;

  data = list->tail->data;
  list_remove(list, list->tail->data);
  
  ENSURE_VALID(list);
  return data;
}

/* -------------------------------------------------------------------------- */
/* returns the number of elements in the list */
unsigned int list_count(linked_list list){
  ENSURE_VALID(list);
  return list->count;
}

/* -------------------------------------------------------------------------- */
/* gets the next item in the list. behaves as if the list is circular. returns NULL when there are no more items in the list
 */
void* list_next(linked_list list){
	list_node* last;
  ENSURE_VALID(list);
  
  if (list->count == 0)
    return NULL;

  if (list->current==NULL)
    list->current = list->root.next;
  
  last = list->current;
  list->current = list->current->next;

  ENSURE_VALID(list);
  return last->data;
}

/* -------------------------------------------------------------------------- */
/* iterate through the list by fcn callback.
   the callback function should return TRUE to continue
   the iteration. or FALSE to end the iteration
   userdata is a user defined convenience value
*/
void list_enumerate_data(linked_list list, listcallback_data callback, void *userdata)
{
	list_node *walker;
  ENSURE_VALID(list);
  
  walker = list->root.next;

  while (walker){
    callback(walker->data, userdata);
    walker = walker->next;
  }
  ENSURE_VALID(list);
}

/* -------------------------------------------------------------------------- */
/* iterate through the list by fcn callback.
   the callback function should return TRUE to continue
   the iteration. or FALSE to end the iteration */
void list_enumerate(linked_list list, listcallback callback)
{
	list_node *walker;
  ENSURE_VALID(list);
  
  walker = list->root.next;

  while (walker){
    callback(walker->data);
    walker = walker->next;
  }
  ENSURE_VALID(list);
}


/*+++EOF++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
